# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright  2014-2022 Vincent Texier <vit@free.fr>
#
# DuniterPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DuniterPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import base64
import binascii
import re
from hashlib import scrypt, sha256
from typing import Optional, Type, Union

import libnacl.sign
import pyaes
from base58 import b58decode, b58encode
from libnacl.utils import load_key

# Scrypt
SCRYPT_PARAMS = {"N": 4096, "r": 16, "p": 1, "seed_length": 32}


class ScryptParams:
    """
    Class to simplify handling of scrypt parameters
    """

    def __init__(
        self,
        n: int = SCRYPT_PARAMS["N"],
        r: int = SCRYPT_PARAMS["r"],
        p: int = SCRYPT_PARAMS["p"],
        seed_length: int = SCRYPT_PARAMS["seed_length"],
    ) -> None:
        """
        Init a ScryptParams instance with crypto parameters

        :param n: scrypt param N, default see constant SCRYPT_PARAMS
        :param r: scrypt param r, default see constant SCRYPT_PARAMS
        :param p: scrypt param p, default see constant SCRYPT_PARAMS
        :param seed_length: scrypt param seed_length, default see constant SCRYPT_PARAMS
        """
        self.N = n
        self.r = r
        self.p = p
        self.seed_length = seed_length


def ensure_bytes(data: Union[str, bytes]) -> bytes:
    """
    Convert data in bytes if data is a string

    :param data: Data
    :return:
    """
    if isinstance(data, str):
        return bytes(data, "utf-8")

    return data


def xor_bytes(b1: bytes, b2: bytes) -> bytearray:
    """
    Apply XOR operation on two bytes arguments

    :param b1: First bytes argument
    :param b2: Second bytes argument
    :return:
    """
    result = bytearray()
    for i1, i2 in zip(b1, b2):
        result.append(i1 ^ i2)
    return result


def hex_decode(data: bytes) -> bytes:
    """
    Convert hexadecimal byte string to bytes

    :param data: Data to decode
    :return:
    """
    return binascii.unhexlify(data)


def hex_encode(data: bytes) -> bytes:
    """
    Convert bytes to hexadecimal byte string

    :param data: Data to encode
    :return:
    """
    return binascii.hexlify(data)


def convert_seedhex_to_seed(seedhex: str) -> bytes:
    """
    Convert seedhex to seed

    :param seedhex: seed coded in hexadecimal base
    :return:
    """
    return bytes(hex_decode(seedhex.encode("utf-8")))


def convert_seed_to_seedhex(seed: bytes) -> str:
    """
    Convert seed to seedhex

    :param seed: seed
    :rtype str:
    """
    return hex_encode(seed).decode("utf-8")


class SigningKey(libnacl.sign.Signer):
    def __init__(self, seed: bytes) -> None:
        """
        Init pubkey property

        :param str seed: Hexadecimal seed string
        """
        super().__init__(seed)
        self.pubkey = b58encode(self.vk).decode("utf8")

    @classmethod
    def from_credentials(
        cls: Type[SigningKey],
        salt: Union[str, bytes],
        password: Union[str, bytes],
        scrypt_params: Optional[ScryptParams] = None,
    ) -> SigningKey:
        """
        Create a SigningKey object from credentials

        :param salt: Secret salt passphrase credential
        :param password: Secret password credential
        :param scrypt_params: ScryptParams instance or None
        """
        if scrypt_params is None:
            scrypt_params = ScryptParams()

        salt = ensure_bytes(salt)
        password = ensure_bytes(password)
        seed = scrypt(
            password,
            salt=salt,
            n=scrypt_params.N,
            r=scrypt_params.r,
            p=scrypt_params.p,
            dklen=scrypt_params.seed_length,
        )

        return cls(seed)

    @classmethod
    def from_credentials_file(
        cls, path: str, scrypt_params: Optional[ScryptParams] = None
    ) -> SigningKey:
        """
        Create a SigningKey object from a credentials file

        A file with the salt on the first line and the password on the second line

        :param path: Credentials file path
        :param scrypt_params: ScryptParams instance or None
        :return:
        """
        # capture credentials from file
        with open(path, encoding="utf-8") as fh:
            lines = fh.readlines()
            assert len(lines) > 1
            salt = lines[0].strip()
            password = lines[1].strip()

        return cls.from_credentials(salt, password, scrypt_params)

    def save_seedhex_file(self, path: str) -> None:
        """
        Save hexadecimal seed file from seed

        :param path: Authentication file path
        """
        seedhex = convert_seed_to_seedhex(self.seed)
        with open(path, "w", encoding="utf-8") as fh:
            fh.write(seedhex)

    @staticmethod
    def from_seedhex_file(path: str) -> SigningKey:
        """
        Return SigningKey instance from Seedhex file

        :param str path: Hexadecimal seed file path
        """
        with open(path, encoding="utf-8") as fh:
            seedhex = fh.read()
        return SigningKey.from_seedhex(seedhex)

    @classmethod
    def from_seedhex(cls: Type[SigningKey], seedhex: str) -> SigningKey:
        """
        Return SigningKey instance from Seedhex

        :param str seedhex: Hexadecimal seed string
        """
        regex_seedhex = re.compile("([0-9a-fA-F]{64})")
        match = re.search(regex_seedhex, seedhex)
        if not match:
            raise Exception("Error: Bad seed hexadecimal format")
        seedhex = match.groups()[0]
        seed = convert_seedhex_to_seed(seedhex)
        return cls(seed)

    def save_private_key(self, path: str) -> None:
        """
        Save authentication file

        :param path: Authentication file path
        """
        self.save(path)

    @staticmethod
    def from_private_key(path: str) -> SigningKey:
        """
        Read authentication file
        Add public key attribute

        :param path: Authentication file path
        """
        key = load_key(path)
        key.pubkey = b58encode(key.vk)
        return key

    def decrypt_seal(self, data: bytes) -> bytes:
        """
        Decrypt bytes data with a curve25519 version of the ed25519 key pair

        :param data: Encrypted data

        :return:
        """
        curve25519_public_key = libnacl.crypto_sign_ed25519_pk_to_curve25519(self.vk)
        curve25519_secret_key = libnacl.crypto_sign_ed25519_sk_to_curve25519(self.sk)
        return libnacl.crypto_box_seal_open(
            data, curve25519_public_key, curve25519_secret_key
        )

    @classmethod
    def from_pubsec_file(cls: Type[SigningKey], path: str) -> SigningKey:
        """
        Return SigningKey instance from Duniter WIF file

        :param path: Path to WIF file
        """
        with open(path, encoding="utf-8") as fh:
            pubsec_content = fh.read()

        # line patterns
        regex_pubkey = re.compile("pub: ([1-9A-HJ-NP-Za-km-z]{43,44})", re.MULTILINE)
        regex_signkey = re.compile("sec: ([1-9A-HJ-NP-Za-km-z]{87,90})", re.MULTILINE)

        # check public key field
        match = re.search(regex_pubkey, pubsec_content)
        if not match:
            raise Exception("Error: Bad format PubSec v1 file, missing public key")

        # check signkey field
        match = re.search(regex_signkey, pubsec_content)
        if not match:
            raise Exception("Error: Bad format PubSec v1 file, missing sec key")

        # capture signkey
        signkey_hex = match.groups()[0]

        # extract seed from signkey
        seed = bytes(b58decode(signkey_hex)[0:32])

        return cls(seed)

    def save_pubsec_file(self, path: str) -> None:
        """
        Save a Duniter PubSec file (PubSec) v1

        :param path: Path to file
        """
        # version
        version = 1

        # base58 encode keys
        base58_signing_key = b58encode(self.sk).decode("utf8")

        # save file
        content = f"Type: PubSec\n\
Version: {version}\n\
pub: {self.pubkey}\n\
sec: {base58_signing_key}"
        with open(path, "w", encoding="utf-8") as fh:
            fh.write(content)

    @classmethod
    def from_wif_or_ewif_file(
        cls, path: str, password: Optional[str] = None
    ) -> SigningKey:
        """
        Return SigningKey instance from Duniter WIF or EWIF file

        :param path: Path to WIF of EWIF file
        :param password: Password needed for EWIF file
        """
        with open(path, encoding="utf-8") as fh:
            wif_content = fh.read()

        # check data field
        regex = re.compile("Data: ([1-9A-HJ-NP-Za-km-z]+)", re.MULTILINE)
        match = re.search(regex, wif_content)
        if not match:
            raise Exception("Error: Bad format WIF or EWIF v1 file")

        # capture hexa wif key
        wif_hex = match.groups()[0]
        return cls.from_wif_or_ewif_hex(wif_hex, password)

    @classmethod
    def from_wif_or_ewif_hex(
        cls, wif_hex: str, password: Optional[str] = None
    ) -> SigningKey:
        """
        Return SigningKey instance from Duniter WIF or EWIF in hexadecimal format

        :param wif_hex: WIF or EWIF string in hexadecimal format
        :param password: Password of EWIF encrypted seed
        """
        wif_bytes = b58decode(wif_hex)

        fi = wif_bytes[0:1]

        if fi == b"\x01":
            result = cls.from_wif_hex(wif_hex)
        elif fi == b"\x02" and password is not None:
            result = cls.from_ewif_hex(wif_hex, password)
        else:
            raise Exception("Error: Bad format: not WIF nor EWIF")

        return result

    @classmethod
    def from_wif_file(cls, path: str) -> SigningKey:
        """
        Return SigningKey instance from Duniter WIF file

        :param path: Path to WIF file
        """
        with open(path, encoding="utf-8") as fh:
            wif_content = fh.read()

        # check data field
        regex = re.compile("Data: ([1-9A-HJ-NP-Za-km-z]+)", re.MULTILINE)
        match = re.search(regex, wif_content)
        if not match:
            raise Exception("Error: Bad format WIF v1 file")

        # capture hexa wif key
        wif_hex = match.groups()[0]
        return cls.from_wif_hex(wif_hex)

    @classmethod
    def from_wif_hex(cls: Type[SigningKey], wif_hex: str) -> SigningKey:
        """
        Return SigningKey instance from Duniter WIF in hexadecimal format

        :param wif_hex: WIF string in hexadecimal format
        """
        wif_bytes = b58decode(wif_hex)
        if len(wif_bytes) != 35:
            raise Exception("Error: the size of WIF is invalid")

        # extract data
        checksum_from_wif = wif_bytes[-2:]
        fi = wif_bytes[0:1]
        seed = wif_bytes[1:-2]
        seed_fi = wif_bytes[0:-2]

        # check WIF format flag
        if fi != b"\x01":
            raise Exception("Error: bad format version, not WIF")

        # checksum control
        checksum = libnacl.crypto_hash_sha256(libnacl.crypto_hash_sha256(seed_fi))[0:2]
        if checksum_from_wif != checksum:
            raise Exception("Error: bad checksum of the WIF")

        return cls(seed)

    def save_wif_file(self, path: str) -> None:
        """
        Save a Wallet Import Format file (WIF) v1

        :param path: Path to file
        """
        # version
        version = 1

        # add format to seed (1=WIF,2=EWIF)
        seed_fi = b"\x01" + self.seed

        # calculate checksum
        sha256_v1 = libnacl.crypto_hash_sha256(seed_fi)
        sha256_v2 = libnacl.crypto_hash_sha256(sha256_v1)
        checksum = sha256_v2[0:2]

        # base58 encode key and checksum
        wif_key = b58encode(seed_fi + checksum).decode("utf8")

        content = f"Type: WIF\n\
Version: {version}\n\
Data: {wif_key}"
        with open(path, "w", encoding="utf-8") as fh:
            fh.write(content)

    @classmethod
    def from_ewif_file(cls, path: str, password: str) -> SigningKey:
        """
        Return SigningKey instance from Duniter EWIF file

        :param path: Path to EWIF file
        :param password: Password of the encrypted seed
        """
        with open(path, encoding="utf-8") as fh:
            wif_content = fh.read()

        # check data field
        regex = re.compile("Data: ([1-9A-HJ-NP-Za-km-z]+)", re.MULTILINE)
        match = re.search(regex, wif_content)
        if not match:
            raise Exception("Error: Bad format EWIF v1 file")

        # capture ewif key
        ewif_hex = match.groups()[0]
        return cls.from_ewif_hex(ewif_hex, password)

    @classmethod
    def from_ewif_hex(
        cls: Type[SigningKey], ewif_hex: str, password: str
    ) -> SigningKey:
        """
        Return SigningKey instance from Duniter EWIF in hexadecimal format

        :param ewif_hex: EWIF string in hexadecimal format
        :param password: Password of the encrypted seed
        """
        ewif_bytes = b58decode(ewif_hex)
        if len(ewif_bytes) != 39:
            raise Exception("Error: the size of EWIF is invalid")

        # extract data
        fi = ewif_bytes[0:1]
        checksum_from_ewif = ewif_bytes[-2:]
        ewif_no_checksum = ewif_bytes[0:-2]
        salt = ewif_bytes[1:5]
        encryptedhalf1 = ewif_bytes[5:21]
        encryptedhalf2 = ewif_bytes[21:37]

        # check format flag
        if fi != b"\x02":
            raise Exception("Error: bad format version, not EWIF")

        # checksum control
        checksum = libnacl.crypto_hash_sha256(
            libnacl.crypto_hash_sha256(ewif_no_checksum)
        )[0:2]
        if checksum_from_ewif != checksum:
            raise Exception("Error: bad checksum of the EWIF")

        # SCRYPT
        password_bytes = password.encode("utf-8")
        scrypt_seed = scrypt(password_bytes, salt=salt, n=16384, r=8, p=8, dklen=64)
        derivedhalf1 = scrypt_seed[0:32]
        derivedhalf2 = scrypt_seed[32:64]

        # AES
        aes = pyaes.AESModeOfOperationECB(derivedhalf2)
        decryptedhalf1 = aes.decrypt(encryptedhalf1)
        decryptedhalf2 = aes.decrypt(encryptedhalf2)

        # XOR
        seed1 = xor_bytes(decryptedhalf1, derivedhalf1[0:16])
        seed2 = xor_bytes(decryptedhalf2, derivedhalf1[16:32])
        seed = bytes(seed1 + seed2)

        # Password Control
        signer = SigningKey(seed)
        salt_from_seed = libnacl.crypto_hash_sha256(
            libnacl.crypto_hash_sha256(b58decode(signer.pubkey))
        )[0:4]
        if salt_from_seed != salt:
            raise Exception("Error: bad Password of EWIF address")

        return cls(seed)

    def save_ewif_file(self, path: str, password: str) -> None:
        """
        Save an Encrypted Wallet Import Format file (WIF v2)

        :param path: Path to file
        :param password:
        """
        # version
        version = 1

        # add version to seed
        salt = libnacl.crypto_hash_sha256(
            libnacl.crypto_hash_sha256(b58decode(self.pubkey))
        )[0:4]

        # SCRYPT
        password_bytes = password.encode("utf-8")
        scrypt_seed = scrypt(password_bytes, salt=salt, n=16384, r=8, p=8, dklen=64)
        derivedhalf1 = scrypt_seed[0:32]
        derivedhalf2 = scrypt_seed[32:64]

        # XOR
        seed1_xor_derivedhalf1_1 = bytes(xor_bytes(self.seed[0:16], derivedhalf1[0:16]))
        seed2_xor_derivedhalf1_2 = bytes(
            xor_bytes(self.seed[16:32], derivedhalf1[16:32])
        )

        # AES
        aes = pyaes.AESModeOfOperationECB(derivedhalf2)
        encryptedhalf1 = aes.encrypt(seed1_xor_derivedhalf1_1)
        encryptedhalf2 = aes.encrypt(seed2_xor_derivedhalf1_2)

        # add format to final seed (1=WIF,2=EWIF)
        seed_bytes = b"\x02" + salt + encryptedhalf1 + encryptedhalf2

        # calculate checksum
        sha256_v1 = libnacl.crypto_hash_sha256(seed_bytes)
        sha256_v2 = libnacl.crypto_hash_sha256(sha256_v1)
        checksum = sha256_v2[0:2]

        # B58 encode final key string
        ewif_key = b58encode(seed_bytes + checksum).decode("utf8")

        # save file
        content = f"Type: EWIF\n\
Version: {version}\n\
Data: {ewif_key}"
        with open(path, "w", encoding="utf-8") as fh:
            fh.write(content)

    @classmethod
    def from_ssb_file(cls: Type[SigningKey], path: str) -> SigningKey:
        """
        Return SigningKey instance from ScuttleButt .ssb/secret file

        :param path: Path to Scuttlebutt secret file
        """
        with open(path, encoding="utf-8") as fh:
            ssb_content = fh.read()

        # check data field
        regex = re.compile(
            '{\\s*"curve": "ed25519",\\s*"public": "(.+)\\.ed25519",\\s*"private":\\s*"(.+)\\.ed25519",\\s*"id":\\s*"@\\1.ed25519"\\s*}',
            re.MULTILINE,
        )
        match = re.search(regex, ssb_content)
        if not match:
            raise Exception("Error: Bad scuttlebutt secret file")

        # capture ssb secret key
        secret = match.groups()[1]

        # extract seed from secret
        seed = bytes(base64.b64decode(secret)[0:32])

        return cls(seed)

    @classmethod
    def from_dubp_mnemonic(
        cls, mnemonic: str, scrypt_params: Optional[ScryptParams] = None
    ) -> SigningKey:
        """
        Generate key pair instance from a DUBP mnemonic passphrase

        See https://git.duniter.org/documents/rfcs/blob/dubp-mnemonic/rfc/0014_Dubp_Mnemonic.md

        :param mnemonic: Passphrase generated from a mnemonic algorithm
        :param scrypt_params: ScryptParams instance (default=None)
        :return:
        """
        if scrypt_params is None:
            scrypt_params = ScryptParams()

        _password = mnemonic.encode("utf-8")  # type: bytes
        _salt = sha256(b"dubp" + _password).digest()  # type: bytes
        _seed = scrypt(
            password=_password,
            salt=_salt,
            n=scrypt_params.N,  # 4096
            r=scrypt_params.r,  # 16
            p=scrypt_params.p,  # 1
            dklen=scrypt_params.seed_length,  # 32
        )  # type: bytes
        return cls(_seed)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Python Substrate Interface Library
#
# Copyright 2018-2023 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import hashlib
import re
import secrets
from hashlib import blake2b
from math import ceil
from typing import Optional, Union

import ed25519_zebra
import nacl.bindings
import nacl.public
import sr25519
from bip39 import bip39_generate, bip39_to_mini_secret, bip39_validate
from eth_keys.datatypes import PrivateKey
from nacl.bindings import (
    crypto_sign_seed_keypair,
)
from scalecodec.base import ScaleBytes
from scalecodec.types import Bytes
from scalecodec.utils.ss58 import ss58_decode, ss58_encode

__all__ = ["DeriveJunction", "Keypair", "KeypairType", "MnemonicLanguageCode"]

RE_JUNCTION = r"(\/\/?)([^/]+)"
JUNCTION_ID_LEN = 32
DEV_PHRASE = "bottom drive obey lake curtain smoke basket hold race lonely fit walk"


class DeriveJunction:
    """
    DeriveJunction class
    """

    def __init__(self, chain_code, is_hard=False):
        """
        Init DeriveJunction instance

        :param chain_code:
        :param is_hard:
        """
        self.chain_code = chain_code
        self.is_hard = is_hard

    @classmethod
    def from_derive_path(cls, path: str, is_hard=False):
        """
        Return DeriveJunction instance from derive path

        :param path: Derive path
        :param is_hard: Hard derivation if True
        :return:
        """
        if path.isnumeric():
            byte_length = ceil(int(path).bit_length() / 8)
            chain_code = int(path).to_bytes(byte_length, "little").ljust(32, b"\x00")

        else:
            path_scale = Bytes()
            path_scale.encode(path)

            if len(path_scale.data) > JUNCTION_ID_LEN:
                chain_code = blake2b(path_scale.data.data, digest_size=32).digest()
            else:
                chain_code = bytes(path_scale.data.data.ljust(32, b"\x00"))

        return cls(chain_code=chain_code, is_hard=is_hard)


def extract_derive_path(derive_path: str):
    """
    Extract derivation path

    :param derive_path: Derive path
    :return:
    """
    path_check = ""
    junctions = []
    paths = re.findall(RE_JUNCTION, derive_path)

    if paths:
        path_check = "".join("".join(path) for path in paths)

        for path_separator, path_value in paths:
            junctions.append(
                DeriveJunction.from_derive_path(
                    path=path_value, is_hard=path_separator == "//"
                )
            )

    if path_check != derive_path:
        raise ValueError(
            'Reconstructed path "{}" does not match input'.format(path_check)
        )

    return junctions


def scale_encode_string(string: str) -> bytes:
    """
    Encode a string to SCALE

    :param string: String to encode
    :return:
    """
    string_in_bytes = string.encode("utf-8")
    length = len(string_in_bytes)

    # Compact encoding pour longueur < 64 (mode single-byte)
    if length < 64:
        encoded_len = (length << 2).to_bytes(1, "little")
    else:
        raise NotImplementedError(
            "SCALE compact encoding only implemented for len < 64"
        )

    return encoded_len + string_in_bytes


def derive_ed25519_hard_junction(seed: bytes, cc: bytes) -> bytes:
    """
    Derive a hard junction

    Same as rust function:

        use codec::Encode;
        ("Ed25519HDKD", secret_seed, cc).using_encoded(sp_crypto_hashing::blake2_256)

    :param seed: Mini-secret seed
    :param cc: Chain code of path junction
    :return:
    """
    assert len(seed) == 32
    assert len(cc) == 32

    encoded = scale_encode_string("Ed25519HDKD") + seed + cc
    return hashlib.blake2b(encoded, digest_size=32).digest()


class KeypairType:
    """
    Type of cryptography, used in `Keypair` instance to encrypt and sign data

    * ED25519 = 0
    * SR25519 = 1
    * ECDSA = 2

    """

    ED25519 = 0
    SR25519 = 1
    ECDSA = 2


class MnemonicLanguageCode:
    """
    Available language codes to generate mnemonics

    * ENGLISH = 'en'
    * CHINESE_SIMPLIFIED = 'zh-hans'
    * CHINESE_TRADITIONAL = 'zh-hant'
    * FRENCH = 'fr'
    * ITALIAN = 'it'
    * JAPANESE = 'ja'
    * KOREAN = 'ko'
    * SPANISH = 'es'

    """

    ENGLISH = "en"
    CHINESE_SIMPLIFIED = "zh-hans"
    CHINESE_TRADITIONAL = "zh-hant"
    FRENCH = "fr"
    ITALIAN = "it"
    JAPANESE = "ja"
    KOREAN = "ko"
    SPANISH = "es"


class Keypair:
    def __init__(
        self,
        ss58_address: Optional[str] = None,
        public_key: Optional[Union[bytes, str]] = None,
        private_key: Optional[Union[bytes, str]] = None,
        ss58_format: Optional[int] = None,
        seed_hex: Optional[Union[str, bytes]] = None,
        crypto_type: int = KeypairType.SR25519,
    ):
        """
        Allows generation of Keypairs from a variety of input combination, such as a public/private key combination,
        mnemonic or URI containing soft and hard derivation paths. With these Keypairs data can be signed and verified

        Parameters
        ----------
        ss58_address: Substrate address
        public_key: hex string or bytes of public_key key
        private_key: hex string or bytes of private key
        ss58_format: Substrate address format, default to 42 when omitted
        seed_hex: hex string of seed
        crypto_type: Use KeypairType.SR25519 or KeypairType.ED25519 cryptography for generating the Keypair
        """

        self.crypto_type = crypto_type
        self.seed_hex = seed_hex
        self.derive_path = None

        if crypto_type != KeypairType.ECDSA and ss58_address and not public_key:
            public_key = ss58_decode(ss58_address, valid_ss58_format=ss58_format)

        if private_key:

            if type(private_key) is str:
                private_key = bytes.fromhex(private_key.replace("0x", ""))  # type: ignore

            if self.crypto_type == KeypairType.SR25519:
                if len(private_key) != 64:
                    raise ValueError("Secret key should be 64 bytes long")
                if not public_key:
                    public_key = sr25519.public_from_secret_key(private_key)

            if self.crypto_type == KeypairType.ECDSA:
                private_key_obj = PrivateKey(private_key)  # type: ignore
                public_key = private_key_obj.public_key.to_address()
                ss58_address = private_key_obj.public_key.to_checksum_address()

        if not public_key:
            raise ValueError("No SS58 formatted address or public key provided")

        if type(public_key) is str:
            public_key = bytes.fromhex(public_key.replace("0x", ""))  # type: ignore

        if crypto_type == KeypairType.ECDSA:
            if len(public_key) != 20:
                raise ValueError("Public key should be 20 bytes long")
        else:
            if len(public_key) != 32:
                raise ValueError("Public key should be 32 bytes long")

            if not ss58_address:
                ss58_address = ss58_encode(public_key, ss58_format=ss58_format)

        self.ss58_format: int = ss58_format  # type: ignore

        self.public_key: bytes = public_key  # type: ignore

        self.ss58_address: str = ss58_address  # type: ignore

        self.private_key: bytes = private_key  # type: ignore

        self.mnemonic = None

    @classmethod
    def generate_mnemonic(
        cls, words: int = 12, language_code: str = MnemonicLanguageCode.ENGLISH
    ) -> str:
        """
        Generates a new seed phrase with given amount of words (default 12)

        Parameters
        ----------
        words: The amount of words to generate, valid values are 12, 15, 18, 21 and 24
        language_code: The language to use, valid values are: 'en', 'zh-hans', 'zh-hant', 'fr', 'it', 'ja', 'ko', 'es'.
                       Defaults to `MnemonicLanguageCode.ENGLISH`

        Returns
        -------
        str: Seed phrase
        """
        return bip39_generate(words, language_code)

    @classmethod
    def validate_mnemonic(
        cls, mnemonic: str, language_code: str = MnemonicLanguageCode.ENGLISH
    ) -> bool:
        """
        Verify if specified mnemonic is valid

        Parameters
        ----------
        mnemonic: Seed phrase
        language_code: The language to use, valid values are: 'en', 'zh-hans', 'zh-hant', 'fr', 'it', 'ja', 'ko', 'es'.
         Defaults to `MnemonicLanguageCode.ENGLISH`

        Returns
        -------
        bool
        """
        return bip39_validate(mnemonic, language_code)

    @classmethod
    def create_from_mnemonic(
        cls,
        mnemonic: str,
        ss58_format: int = 42,
        crypto_type: int = KeypairType.SR25519,
        language_code: str = MnemonicLanguageCode.ENGLISH,
    ) -> "Keypair":
        """
        Create a Keypair for given memonic

        Parameters
        ----------
        mnemonic: Seed phrase
        ss58_format: Substrate address format
        crypto_type: Use `KeypairType.SR25519` or `KeypairType.ED25519` cryptography for generating the Keypair
        language_code: The language to use, valid values are: 'en', 'zh-hans', 'zh-hant', 'fr', 'it', 'ja', 'ko', 'es'.
         Defaults to `MnemonicLanguageCode.ENGLISH`

        Returns
        -------
        Keypair
        """

        if crypto_type == KeypairType.ECDSA:
            raise ValueError("ECDSA not supported")
            # if language_code != MnemonicLanguageCode.ENGLISH:
            #     raise ValueError("ECDSA mnemonic only supports english")
            #
            # private_key = mnemonic_to_ecdsa_private_key(mnemonic)
            # keypair = cls.create_from_private_key(private_key, ss58_format=ss58_format, crypto_type=crypto_type)

        else:
            seed_array = bip39_to_mini_secret(mnemonic, "", language_code)

            keypair = cls.create_from_seed(
                seed_hex=binascii.hexlify(bytearray(seed_array)).decode("ascii"),
                ss58_format=ss58_format,
                crypto_type=crypto_type,
            )

        keypair.mnemonic = mnemonic  # type: ignore

        return keypair

    @classmethod
    def create_from_seed(
        cls,
        seed_hex: Union[bytes, str],
        ss58_format: int = 42,
        crypto_type: int = KeypairType.SR25519,
    ) -> "Keypair":
        """
        Create a Keypair for given seed

        Parameters
        ----------
        seed_hex: hex string of seed
        ss58_format: Substrate address format
        crypto_type: Use KeypairType.SR25519 or KeypairType.ED25519 cryptography for generating the Keypair

        Returns
        -------
        Keypair
        """

        if type(seed_hex) is str:
            seed_hex = bytes.fromhex(seed_hex.replace("0x", ""))  # type: ignore

        if crypto_type == KeypairType.SR25519:
            public_key, private_key = sr25519.pair_from_seed(seed_hex)
        elif crypto_type == KeypairType.ED25519:
            private_key, public_key = ed25519_zebra.ed_from_seed(seed_hex)
        else:
            raise ValueError('crypto_type "{}" not supported'.format(crypto_type))

        ss58_address = ss58_encode(public_key, ss58_format)

        return cls(
            ss58_address=ss58_address,
            public_key=public_key,
            private_key=private_key,
            ss58_format=ss58_format,
            crypto_type=crypto_type,
            seed_hex=seed_hex,
        )

    @classmethod
    def create_from_uri(
        cls,
        suri: str,
        ss58_format: int = 42,
        crypto_type: int = KeypairType.SR25519,
        language_code: str = MnemonicLanguageCode.ENGLISH,
    ) -> "Keypair":
        """
        Creates Keypair for specified suri in format: `[mnemonic]/[soft-path]//[hard-path]`

        Parameters
        ----------
        suri: str
            Secret URI including mnemonic and derivation path.
        ss58_format: int
            Substrate address format
        crypto_type: int
            Use KeypairType.SR25519 or KeypairType.ED25519
        language_code: MnemonicLanguageCode
            Language of the mnemonic phrase

        Returns
        -------
        Keypair
        """

        if suri and suri.startswith("/"):
            suri = DEV_PHRASE + suri

        suri_regex = re.match(
            r"^(?P<phrase>.[^/]+( .[^/]+)*)(?P<path>(//?[^/]+)*)(///(?P<password>.*))?$",
            suri,
        )

        suri_parts = suri_regex.groupdict()  # type: ignore

        if crypto_type == KeypairType.ECDSA:
            raise ValueError("ECDSA not supported")

        if suri_parts["password"]:
            raise NotImplementedError(
                f"Passwords in suri not supported for crypto_type '{crypto_type}'"
            )

        derived_keypair = cls.create_from_mnemonic(
            suri_parts["phrase"],
            ss58_format=ss58_format,
            crypto_type=crypto_type,
            language_code=language_code,
        )

        if suri_parts["path"] != "":
            derived_keypair.derive_path = suri_parts["path"]  # type: ignore
            derive_junctions = extract_derive_path(suri_parts["path"])

            child_pubkey = derived_keypair.public_key
            child_privkey = derived_keypair.private_key
            seed = derived_keypair.seed_hex
            assert isinstance(seed, bytes)

            for junction in derive_junctions:
                if crypto_type == KeypairType.SR25519:
                    if junction.is_hard:
                        _, child_pubkey, child_privkey = sr25519.hard_derive_keypair(
                            (junction.chain_code, child_pubkey, child_privkey), b""
                        )
                    else:
                        _, child_pubkey, child_privkey = sr25519.derive_keypair(
                            (junction.chain_code, child_pubkey, child_privkey), b""
                        )

                elif crypto_type == KeypairType.ED25519:
                    if not junction.is_hard:
                        raise NotImplementedError(
                            "Soft derivation not supported for ED25519"
                        )
                    # get derived seed from parent seed and chaincode
                    seed = derive_ed25519_hard_junction(seed, junction.chain_code)

                    # generate keypair
                    child_pubkey, child_privkey = crypto_sign_seed_keypair(seed)

                else:
                    raise NotImplementedError(
                        f"Derivation for crypto type {crypto_type} not supported"
                    )

            derived_keypair = Keypair(
                public_key=child_pubkey,
                private_key=child_privkey,
                ss58_format=ss58_format,
                crypto_type=crypto_type,
            )

        return derived_keypair

    @classmethod
    def create_from_private_key(
        cls,
        private_key: Union[bytes, str],
        public_key: Optional[Union[bytes, str]] = None,
        ss58_address: Optional[str] = None,
        ss58_format: Optional[int] = None,
        crypto_type: int = KeypairType.SR25519,
    ) -> "Keypair":
        """
        Creates Keypair for specified public/private keys
        Parameters
        ----------
        private_key: hex string or bytes of private key
        public_key: hex string or bytes of public key
        ss58_address: Substrate address
        ss58_format: Substrate address format, default = 42
        crypto_type: Use KeypairType.[SR25519|ED25519|ECDSA] cryptography for generating the Keypair

        Returns
        -------
        Keypair
        """

        return cls(
            ss58_address=ss58_address,
            public_key=public_key,
            private_key=private_key,
            ss58_format=ss58_format,
            crypto_type=crypto_type,
        )

    # @classmethod
    # def create_from_encrypted_json(cls, json_data: Union[str, dict], passphrase: str,
    #                                ss58_format: int = None) -> 'Keypair':
    #     """
    #     Create a Keypair from a PolkadotJS format encrypted JSON file
    #
    #     Parameters
    #     ----------
    #     json_data: Dict or JSON string containing PolkadotJS export format
    #     passphrase: Used to encrypt the keypair
    #     ss58_format: Which network ID to use to format the SS58 address (42 for testnet)
    #
    #     Returns
    #     -------
    #     Keypair
    #     """
    #
    #     if type(json_data) is str:
    #         json_data = json.loads(json_data)
    #
    #     private_key, public_key = decode_pair_from_encrypted_json(json_data, passphrase)
    #
    #     if 'sr25519' in json_data['encoding']['content']:
    #         crypto_type = KeypairType.SR25519
    #     elif 'ed25519' in json_data['encoding']['content']:
    #         crypto_type = KeypairType.ED25519
    #         # Strip the nonce part of the private key
    #         private_key = private_key[0:32]
    #     else:
    #         raise NotImplementedError("Unknown KeypairType found in JSON")
    #
    #     if ss58_format is None and 'address' in json_data:
    #         ss58_format = get_ss58_format(json_data['address'])
    #
    #     return cls.create_from_private_key(private_key, public_key, ss58_format=ss58_format, crypto_type=crypto_type)
    #
    # def export_to_encrypted_json(self, passphrase: str, name: str = None) -> dict:
    #     """
    #     Export Keypair to PolkadotJS format encrypted JSON file
    #
    #     Parameters
    #     ----------
    #     passphrase: Used to encrypt the keypair
    #     name: Display name of Keypair used
    #
    #     Returns
    #     -------
    #     dict
    #     """
    #     if not name:
    #         name = self.ss58_address
    #
    #     if self.crypto_type != KeypairType.SR25519:
    #         raise NotImplementedError(f"Cannot create JSON for crypto_type '{self.crypto_type}'")
    #
    #     # Secret key from PolkadotJS is an Ed25519 expanded secret key, so has to be converted
    #     # https://github.com/polkadot-js/wasm/blob/master/packages/wasm-crypto/src/rs/sr25519.rs#L125
    #     converted_private_key = sr25519.convert_secret_key_to_ed25519(self.private_key)
    #
    #     encoded = encode_pair(self.public_key, converted_private_key, passphrase)
    #
    #     json_data = {
    #         "encoded": b64encode(encoded).decode(),
    #         "encoding": {"content": ["pkcs8", "sr25519"], "type": ["scrypt", "xsalsa20-poly1305"], "version": "3"},
    #         "address": self.ss58_address,
    #         "meta": {
    #             "name": name, "tags": [], "whenCreated": int(time.time())
    #         }
    #     }
    #
    #     return json_data

    def sign(self, data: Union[ScaleBytes, bytes, str]) -> bytes:
        """
        Creates a signature for given data

        Parameters
        ----------
        data: data to sign in `Scalebytes`, bytes or hex string format

        Returns
        -------
        signature in bytes

        """
        if type(data) is ScaleBytes:
            data = bytes(data.data)  # type: ignore
        elif data[0:2] == "0x":
            data = bytes.fromhex(data[2:])  # type: ignore
        elif type(data) is str:
            data = data.encode()  # type: ignore

        if not self.private_key:
            raise ValueError("No private key set to create signatures")

        if self.crypto_type == KeypairType.SR25519:
            signature = sr25519.sign((self.public_key, self.private_key), data)

        elif self.crypto_type == KeypairType.ED25519:
            signature = ed25519_zebra.ed_sign(self.private_key, data)

        elif self.crypto_type == KeypairType.ECDSA:
            raise ValueError("Crypto type not supported")
            # signature = ecdsa_sign(self.private_key, data)

        else:
            raise ValueError("Crypto type not supported")

        return signature

    def verify(
        self, data: Union[ScaleBytes, bytes, str], signature: Union[bytes, str]
    ) -> bool:
        """
        Verifies data with specified signature

        Parameters
        ----------
        data: data to be verified in `Scalebytes`, bytes or hex string format
        signature: signature in bytes or hex string format

        Returns
        -------
        True if data is signed with this Keypair, otherwise False
        """

        if type(data) is ScaleBytes:
            data = bytes(data.data)  # type: ignore
        elif data[0:2] == "0x":
            data = bytes.fromhex(data[2:])  # type: ignore
        elif type(data) is str:
            data = data.encode()  # type: ignore

        if type(signature) is str and signature[0:2] == "0x":
            signature = bytes.fromhex(signature[2:])  # type: ignore

        if type(signature) is not bytes:
            raise TypeError("Signature should be of type bytes or a hex-string")

        if self.crypto_type == KeypairType.SR25519:
            crypto_verify_fn = sr25519.verify
        elif self.crypto_type == KeypairType.ED25519:
            crypto_verify_fn = ed25519_zebra.ed_verify
        elif self.crypto_type == KeypairType.ECDSA:
            raise ValueError("Crypto type not supported")
            # crypto_verify_fn = ecdsa_verify
        else:
            raise ValueError("Crypto type not supported")

        verified = crypto_verify_fn(signature, data, self.public_key)

        if not verified:
            # Another attempt with the data wrapped, as discussed in https://github.com/polkadot-js/extension/pull/743
            # Note: As Python apps are trusted sources on its own, no need to wrap data when signing from this lib
            verified = crypto_verify_fn(
                signature, b"<Bytes>" + data + b"</Bytes>", self.public_key  # type: ignore
            )

        return verified

    def encrypt_message(
        self,
        message: Union[bytes, str],
        recipient_public_key: bytes,
        nonce: bytes = secrets.token_bytes(24),
    ) -> bytes:
        """
        Encrypts message with for specified recipient

        Parameters
        ----------
        message: message to be encrypted, bytes or string
        recipient_public_key: recipient's public key
        nonce: the nonce to use in the encryption

        Returns
        -------
        Encrypted message
        """

        if not self.private_key:
            raise ValueError("No private key set to encrypt")
        if self.crypto_type != KeypairType.ED25519:
            raise ValueError("Only ed25519 keypair type supported")
        curve25519_public_key = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(
            recipient_public_key
        )
        recipient = nacl.public.PublicKey(curve25519_public_key)
        private_key = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(
            self.private_key + self.public_key
        )
        sender = nacl.public.PrivateKey(private_key)
        box = nacl.public.Box(sender, recipient)
        return box.encrypt(
            message if isinstance(message, bytes) else message.encode("utf-8"), nonce
        )

    def decrypt_message(
        self, encrypted_message_with_nonce: bytes, sender_public_key: bytes
    ) -> bytes:
        """
        Decrypts message from a specified sender

        Parameters
        ----------
        encrypted_message_with_nonce: message to be decrypted
        sender_public_key: sender's public key

        Returns
        -------
        Decrypted message
        """

        if not self.private_key:
            raise ValueError("No private key set to decrypt")
        if self.crypto_type != KeypairType.ED25519:
            raise ValueError("Only ed25519 keypair type supported")
        private_key = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(
            self.private_key + self.public_key
        )
        recipient = nacl.public.PrivateKey(private_key)
        curve25519_public_key = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(
            sender_public_key
        )
        sender = nacl.public.PublicKey(curve25519_public_key)
        return nacl.public.Box(recipient, sender).decrypt(encrypted_message_with_nonce)

    def __repr__(self):
        if self.ss58_address:
            return "<Keypair (address={})>".format(self.ss58_address)
        else:
            return "<Keypair (public_key=0x{})>".format(self.public_key.hex())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc

from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.network.connection import ConnectionInterface
from tikka.interfaces.adapters.network.network import NetworkInterface


class ConnectionsInterface(abc.ABC):
    """
    ConnectionsInterface class
    """

    def __init__(
        self, network: NetworkInterface, event_dispatcher: EventDispatcher
    ) -> None:
        """
        Init Connections domain instance

        :param network: NetworkInterface instance
        :param event_dispatcher: EventDispatcher instance
        :return:
        """
        self.network = network
        self.event_dispatcher = event_dispatcher

    @abc.abstractmethod
    def disconnect_all(self):
        """
        Disconnect all connections

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def are_all_connected(self) -> bool:
        """
        Return True if connection is active, False otherwise

        :return:
        """
        raise NotImplementedError

    @property
    def node(self) -> ConnectionInterface:
        """
        Return a ConnectionInterface instance for the node

        :return:
        """
        raise NotImplementedError

    @property
    def indexer(self) -> ConnectionInterface:
        """
        Return a ConnectionInterface instance for the indexer

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import pickle
from typing import Any, Iterable, List, Optional, Tuple, Type, Union
from uuid import UUID

from PyQt5.QtCore import (
    QAbstractItemModel,
    QLocale,
    QMimeData,
    QModelIndex,
    Qt,
    QVariant,
)
from PyQt5.QtGui import QColor, QFont, QPixmap

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.category import Category
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY
from tikka.domains.entities.identity import IdentityStatus
from tikka.domains.entities.smith import SmithStatus
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_IDENTITY_MEMBER,
    ICON_IDENTITY_MEMBER_NOT_VALIDATED,
    ICON_IDENTITY_NOT_MEMBER,
    ICON_SMITH,
    SELECTED_UNIT_PREFERENCES_KEY,
)

WalletLockStatusRole = Qt.UserRole + 1000
BalanceStyledRole = Qt.UserRole + 1001


class TreeItem:
    def __init__(
        self, element: Optional[Union[Category, Account]], parent: Optional[TreeItem]
    ) -> None:
        """
        Init TreeItem instance with element and parent

        :param element: Category or Account instance
        :param parent: TreeItem instance
        """
        self.element = element
        self._children: List[TreeItem] = []
        self._parent = parent
        self._row = 0

    def child_count(self):
        """
        Return number of children

        :return:
        """
        return len(self._children)

    def child(self, row) -> Optional[TreeItem]:
        """
        Return child at row

        :param row: Row number
        :return:
        """
        if 0 <= row < self.child_count():
            return self._children[row]
        return None

    def parent(self) -> Optional[TreeItem]:
        """
        Return parent

        :return:
        """
        return self._parent

    def row(self):
        """
        Return row number in parent's children

        :return:
        """
        return self._row

    def insert_child(self, position: int, child: TreeItem) -> bool:
        """
        Insert child at position row

        :param position: Row to insert child
        :param child: TreeItem instance
        :return:
        """
        self._children.insert(position, child)
        return True

    def add_child(self, child: TreeItem) -> None:
        """
        Add a child to item

        :param child: TreeItem instance
        :return:
        """
        child.set_parent(self)
        child.set_row(len(self._children))
        self._children.append(child)

    def set_parent(self, parent: TreeItem) -> None:
        """
        Set parent item

        :param parent: TreeItem instance
        :return:
        """
        self._parent = parent

    def set_row(self, row: int) -> None:
        """
        Set row value

        :param row: Row value
        :return:
        """
        self._row = row


class AccountTreeModel(QAbstractItemModel):
    """
    AccountTreeModel class that drives the population of tree display
    """

    def __init__(self, application: Application, locale: QLocale):
        """
        Init AccountTreeModel

        :param application: Application instance
        :param locale: QLocale instance
        """
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext

        # get parent widget locale to localize balance display
        self.locale = locale

        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # drag and drop mime-type
        self.mime_type = "application/vnd.text.list"

        # Number of column displayed, see self.data()
        self._column_count = 4

        self.root_item = TreeItem(None, None)
        self.categories: List[Category] = []
        self.accounts: List[Account] = []

        self.display_identity_icon = {
            IdentityStatus.UNCONFIRMED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.UNVALIDATED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.MEMBER.value: QPixmap(ICON_IDENTITY_MEMBER),
            IdentityStatus.NOT_MEMBER.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
            IdentityStatus.REVOKED.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
        }

        self.init_data()

    def init_data(self):
        """
        Fill data from repository

        :return:
        """
        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        self.beginResetModel()
        self.root_item = TreeItem(None, None)
        self.categories = self.application.categories.list_all()
        self.accounts = self.application.accounts.get_list()
        self.recursive_append_children(None, self.root_item)
        self.endResetModel()

    def account_list_by_category_id(self, category_id: Optional[UUID]) -> List[Account]:
        """
        Return all accounts in category_id

        :param category_id: Category ID
        :return:
        """
        result = []
        for account in self.accounts:
            # do not list derived accounts
            if account.category_id is None and account.root is not None:
                continue
            if account.category_id == category_id:
                result.append(account)

        return result

    def category_children(self, category_id: Optional[UUID]) -> List[Category]:
        """
        Return categories with parent category_id

        :param category_id: Parent category ID
        :return:
        """
        return [
            category
            for category in self.categories
            if category.parent_id == category_id
        ]

    def get_derivation_accounts(self, address: str) -> List[Account]:
        """
        Return derived accounts from root address

        :param address: Root account address
        :return:
        """
        return sorted(
            [account for account in self.accounts if account.root == address],
            key=lambda account: account.path,  # type: ignore
        )

    def recursive_append_children(
        self, element: Optional[Union[Category, Account]], item: TreeItem
    ) -> int:
        """
        Add children items from element to item

        :param element: Account or Category instance or None
        :param item: TreeItem instance
        :return:
        """
        children_balance = 0
        if element is None or isinstance(element, Category):
            parent_category_id = None if element is None else element.id

            # list of categories in category
            for category in self.category_children(parent_category_id):
                category_item = TreeItem(category, self.root_item)
                item.add_child(category_item)
                children_balance += self.recursive_append_children(
                    category, category_item
                )

            # list of accounts in category
            for account in self.account_list_by_category_id(parent_category_id):
                account_item = TreeItem(account, item)
                item.add_child(account_item)
                children_balance += self.recursive_append_children(
                    account, account_item
                )
                if account.balance is not None:
                    children_balance += account.balance

        elif isinstance(element, Account):
            # list of derivation from this account
            for account in self.get_derivation_accounts(element.address):
                if account.balance is not None:
                    children_balance += account.balance
                account_item = TreeItem(account, item)
                item.add_child(account_item)

        if isinstance(element, Category):
            element.balance = children_balance

        return children_balance

    def index(
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        """
        Return QModelIndex for row, column and parent

        :param row: Row index
        :param column: Column index
        :param parent: Parent QModelIndex instance
        :return:
        """
        if not parent or not parent.isValid():
            parent_item = self.root_item
        else:
            parent_item = parent.internalPointer()

        if not QAbstractItemModel.hasIndex(self, row, column, parent):
            return QModelIndex()

        child_item = parent_item.child(row)
        if child_item:
            return QAbstractItemModel.createIndex(self, row, column, child_item)

        return QModelIndex()

    def parent(self, child: QModelIndex) -> QModelIndex:  # type: ignore
        """
        Return parent QModelIndex of child QModelIndex

        :param child: QModelIndex instance
        :return:
        """
        if child.isValid():
            parent_item = child.internalPointer().parent()
            if parent_item:
                return QAbstractItemModel.createIndex(
                    self, parent_item.row(), 0, parent_item
                )
        return QModelIndex()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """
        Return row count of parent QModelIndex

        :param parent: QModelIndex instance
        :return:
        """
        if parent.isValid():
            return parent.internalPointer().child_count()
        return self.root_item.child_count()

    def columnCount(
        self, parent: QModelIndex = QModelIndex()  # pylint: disable=unused-argument
    ) -> int:
        """
        Return column count of parent QModelIndex

        :param parent: QModelIndex instance
        :return:
        """
        return self._column_count

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        data = QVariant()
        if not index.isValid():
            return data

        tree_item = index.internalPointer()
        # display account address or name or identity name#index or category name in first column
        # only this column is indented
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.EditRole):
            if isinstance(tree_item.element, Account):
                if tree_item.element.name is None:
                    identity = self.application.identities.get_by_address(
                        tree_item.element.address
                    )
                    if identity is not None:
                        if identity.name is not None:
                            data = QVariant(f"{identity.name}#{identity.index}")
                        else:
                            data = QVariant(f"#{identity.index}")
                else:
                    data = QVariant(tree_item.element.name)
            else:
                # category name
                data = QVariant(tree_item.element.name)

        # display wallet lock status icon for account in first column
        if (
            index.column() == 0
            and role == WalletLockStatusRole
            and isinstance(tree_item.element, Account)
        ):
            if self.application.wallets.exists(tree_item.element.address):
                data = (
                    QVariant(
                        QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED).scaled(
                            16,
                            18,
                            aspectRatioMode=Qt.KeepAspectRatio,
                            transformMode=Qt.SmoothTransformation,
                        )
                    )
                    if self.application.wallets.is_unlocked(tree_item.element.address)
                    else QVariant(
                        QPixmap(ICON_ACCOUNT_WALLET_LOCKED).scaled(
                            16,
                            18,
                            aspectRatioMode=Qt.KeepAspectRatio,
                            transformMode=Qt.SmoothTransformation,
                        )
                    )
                )
            else:
                data = QVariant(
                    QPixmap(ICON_ACCOUNT_NO_WALLET).scaled(
                        16,
                        18,
                        aspectRatioMode=Qt.KeepAspectRatio,
                        transformMode=Qt.SmoothTransformation,
                    )
                )

        # display icon of identity membership status
        if (
            index.column() == 0
            and role == Qt.DecorationRole
            and isinstance(tree_item.element, Account)
        ):
            identity = self.application.identities.get_by_address(
                tree_item.element.address
            )
            if identity is not None:
                data = QVariant(
                    self.display_identity_icon[identity.status.value].scaled(
                        16, 18, aspectRatioMode=Qt.KeepAspectRatio
                    )
                )
                smith = self.application.smiths.get(identity.index)
                if smith is not None and smith.status == SmithStatus.SMITH:
                    data = QVariant(
                        QPixmap(ICON_SMITH).scaled(
                            16, 18, aspectRatioMode=Qt.KeepAspectRatio
                        )
                    )

        # display account address
        if (
            index.column() == 1
            and role == Qt.DisplayRole
            and isinstance(tree_item.element, Account)
        ):
            data = QVariant(tree_item.element.address)
        if (
            index.column() == 1
            and role == Qt.ToolTipRole
            and isinstance(tree_item.element, Account)
            and tree_item.element.crypto_type == AccountCryptoType.ED25519
        ):
            data = QVariant(
                tree_item.element.get_v1_address(
                    self.application.currencies.get_current().ss58_format
                )
            )

        # font style
        if role == Qt.FontRole and isinstance(tree_item.element, Account):
            if index.column() == 1:
                # display addresses in monospace font
                font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
                # if font is not found, choose equivalent monospace font on system
                font.setStyleHint(QFont.Monospace)
                # font.setBold(True)
                # display root account with underline
                if tree_item.element.root is None:
                    font.setUnderline(True)
            else:
                font = QFont()

            return QVariant(font)

        # display account derivation path
        if (
            index.column() == 2
            and role == Qt.DisplayRole
            and isinstance(tree_item.element, Account)
        ):
            if tree_item.element.path is not None:
                data = QVariant(tree_item.element.path)

        # display account balance
        if (
            index.column() == 3
            and role == BalanceStyledRole
            and tree_item.element.balance is not None
        ):
            data = QVariant(
                self.locale.toCurrencyString(
                    self.amount.value(tree_item.element.balance), self.amount.symbol()
                )
            )

        if role == Qt.TextColorRole and index.column() == 0:
            if isinstance(tree_item.element, Account):
                if tree_item.element.name is None:
                    identity = self.application.identities.get_by_address(
                        tree_item.element.address
                    )
                    if identity is not None:
                        data = QVariant(QColor("blue"))

        return data

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """
        Return Qt.ItemFlags for index

        :param index: QModelIndex instance
        :return:
        """
        flags = super().flags(index)
        tree_item: TreeItem = index.internalPointer()

        if index.column() == 0:
            if isinstance(tree_item.element, Category):
                # category name is editable, can be dragged and dropped on
                return (
                    flags
                    | Qt.ItemIsEditable
                    | Qt.ItemIsDragEnabled
                    | Qt.ItemIsDropEnabled
                )
            # account can be dragged and dropped on
            return flags | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled

        return flags

    def setData(self, index: QModelIndex, value: Any, role: int = 0) -> bool:
        """
        Set data after treeView editing

        :param index: QModelIndex instance
        :param value: Editor Value
        :param role: Role flag
        :return:
        """
        super().setData(index, role)

        tree_item: TreeItem = index.internalPointer()
        # if category name edited (not dropped where value is None)...
        if (
            role == Qt.EditRole
            and index.column() == 0
            and isinstance(tree_item.element, Category)
            and value is not None
        ):
            # update name
            tree_item.element.name = value
            # update repository
            self.application.categories.update(tree_item.element)

        self.dataChanged.emit(index, index)
        return True

    def insertRow(self, row: int, parent: QModelIndex = QModelIndex()) -> bool:
        """
        Insert row in parent index

        :param row: Index of row from 0 in parent children
        :param parent: QModelIndex instance
        :return:
        """
        self.beginInsertRows(parent, row, row)
        parent_item: TreeItem = parent.internalPointer()
        parent_id = None
        if parent_item is None:
            parent_item = self.root_item
        elif isinstance(parent_item.element, Category):
            parent_id = parent_item.element.id
        category = self.application.categories.create("", parent_id)
        success = parent_item.insert_child(row, TreeItem(category, parent_item))
        self.application.categories.repository.add(category)
        self.endInsertRows()
        return success

    def supportedDropActions(self):
        """
        Return supported drop action flags

        :return:
        """
        return Qt.MoveAction

    def mimeTypes(self) -> List[str]:
        """
        Return supported mime types for drag and drop

        :return:
        """
        return [self.mime_type]

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        """
        Return encoded mime-type values for indexes

        :param indexes: List of indexes
        :return:
        """
        sorted_indices = sorted(
            [index for index in indexes if index.isValid()],
            key=lambda index: index.row(),
        )

        data_list: List[Tuple[Union[Type[Category], Type[Account]], str]] = []
        for index in sorted_indices:
            if not index.isValid():
                continue
            if index.column() == 0:
                tree_item: TreeItem = index.internalPointer()
                if tree_item is None:
                    continue
                if isinstance(tree_item.element, Category):
                    data_list.append(
                        (tree_item.element.__class__, tree_item.element.id.hex)
                    )
                elif isinstance(tree_item.element, Account):
                    data_list.append(
                        (tree_item.element.__class__, tree_item.element.address)
                    )

        encoded_data = pickle.dumps(data_list)
        mime_data = QMimeData()
        mime_data.setData(self.mime_type, encoded_data)
        return mime_data

    def dropMimeData(
        self,
        data: QMimeData,
        action: Qt.DropAction,
        row: int,  # pylint: disable=unused-argument
        column: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Handle dropped data

        :param data: QMimeData instance
        :param action: Action flag
        :param row: Row dropped on
        :param column: Column dropped on
        :param parent: QModelIndex instance
        :return:
        """
        if action == Qt.IgnoreAction:
            return True
        if not data.hasFormat(self.mime_type) or column > 0 or not parent.isValid():
            return False

        encoded_data = data.data(self.mime_type)
        decoded_data = pickle.loads(bytes(encoded_data))
        parent_tree_item: TreeItem = parent.internalPointer()

        # if dropped on no item or on item without element...
        if parent_tree_item is None or parent_tree_item.element is None:
            return False

        parent_category_id = None

        # if dropped on a category...
        if isinstance(parent_tree_item.element, Category):
            parent_category_id = parent_tree_item.element.id
        elif isinstance(parent_tree_item.element, Account):
            # if dropped on a derived account...
            if parent_tree_item.element.root is not None:
                # get root account category
                root_account = self.application.accounts.get_by_address(
                    parent_tree_item.element.root
                )
                if root_account is not None:
                    parent_category_id = root_account.category_id
                else:
                    return False
            else:
                # get root account category
                root_account = self.application.accounts.get_by_address(
                    parent_tree_item.element.address
                )
                if root_account is not None:
                    parent_category_id = root_account.category_id
                else:
                    return False

        for element_data in decoded_data:
            element_class, element_pk = element_data
            if element_class == Category:
                category = self.application.categories.get(UUID(hex=element_pk))
                if category is not None:
                    category.parent_id = parent_category_id
                    self.application.categories.repository.update(category)
            else:
                account = self.application.accounts.get_by_address(element_pk)
                if account is not None:
                    # if root account...
                    if account.root is None:
                        account.category_id = parent_category_id
                        self.application.accounts.repository.update(account)
                    else:
                        # get root account
                        root_account = self.application.accounts.get_by_address(
                            account.root
                        )
                        if root_account is not None:
                            root_account.category_id = parent_category_id
                            self.application.accounts.repository.update(root_account)
        return True



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtGui import QFont, QFontDatabase
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.entities.constants import (
    NUMERIC_DISPLAY_COLOR_GREEN,
    NUMERIC_DISPLAY_COLOR_RED,
)
from tikka.slots.pyqt.resources.gui.widgets.account_transfer_row_rc import (
    Ui_AccountTransferRowWidget,
)


class AccountTransferRowWidget(QWidget, Ui_AccountTransferRowWidget):
    """
    AccountTransferRowWidget class
    """

    def __init__(
        self,
        account_address: str,
        account_name: str,
        identity_name: str,
        account_has_identity: bool,
        datetime: str,
        amount: str,
        comment: Optional[str] = None,
        comment_type: Optional[str] = None,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountTransferRowWidget instance

        :param account_address: Account address label text
        :param account_name: Account name
        :param identity_name: Identity name label text
        :param account_has_identity: True if account name is identity name
        :param datetime: Datetime label text
        :param amount: Amount label text
        :param comment: Optional comment label text, default to None
        :param comment_type: Optional comment type, default to None
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        account_identity_name_color = "blue"
        amount_received_background_color = NUMERIC_DISPLAY_COLOR_GREEN
        amount_paid_background_color = NUMERIC_DISPLAY_COLOR_RED
        amount_color = "white"

        self.accountAddressValuelabel.setText(account_address)

        if account_has_identity:
            self.identityNameValueLabel.setStyleSheet(
                f'color: "{account_identity_name_color}"'
            )
            self.accountNameValueLabel.setText(
                f" - {account_name}" if account_name is not None else ""
            )
            self.identityNameValueLabel.setText(identity_name)
        else:
            self.identityNameValueLabel.setText(account_name)
            self.accountNameValueLabel.setText("")

        self.datetimeValueLabel.setText(datetime)

        # set amount style
        if amount[0] == "-":
            amount_background = amount_paid_background_color
        else:
            amount_background = amount_received_background_color

        self.amountValueLabel.setText(amount)
        self.amountValueLabel.setStyleSheet(
            f""" 
            color: {amount_color};
            border: 3px solid {amount_background};
            border-radius: 10%;
            background: {amount_background};
            """
        )

        if comment is None:
            self.commentValueLabel.setText("")
        else:
            if comment_type == "UNICODE":
                fonts = QFontDatabase()
                names = fonts.families()
                if "Symbola" in names:
                    self.commentValueLabel.setFont(QFont("Symbola"))
                elif "Emoji One" in names:
                    self.commentValueLabel.setFont(QFont("Emoji One"))

            # comment = "0 " * 73
            self.commentValueLabel.setText(comment)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    account_address = "5D3r...8uI9"
    account_name = "Vince"
    identity_name = "vic#67"
    datetime = "01/02/2025 - 21:50:06"
    amount = "-257 G1"
    comment = "Merci pour les crêpes"
    widget = AccountTransferRowWidget(
        account_address,
        account_name,
        identity_name,
        True,
        datetime,
        amount,
        comment,
        "UNICODE",
        main_window,
    )
    main_window.setCentralWidget(widget)

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re
from typing import Optional

from tikka.domains.entities.constants import DERIVATION_PATH_MEMBER
from tikka.libs.keypair import Keypair

RE_TRANSPARENT_DERIVATION_PATH_PATTERN = re.compile(r"^\/\/(\d+)$")


def get_root_address_from_suri(
    suri: str, crypto_type: int, ss58_format: int, language_code: str
) -> str:
    """
    Return the root address of a derivation path from a Substrate URI

    :param suri: Substrate URI
    :param crypto_type: KeypairType constant
    :param ss58_format: Blockchain address format number
    :param language_code: Mnemonic language code as Keypair.MnemonicLanguageCode constant
    :return:
    """
    parts = re.split("/", suri)
    mnemonic = parts[0]
    keypair = Keypair.create_from_mnemonic(
        mnemonic,
        crypto_type=crypto_type,
        ss58_format=ss58_format,
        language_code=language_code,
    )

    return keypair.ss58_address


def get_path_from_suri(suri: str) -> Optional[str]:
    """
    Return the derivation path from a Substrate URI

    :param suri: Substrate URI
    :return:
    """
    parts = re.split("/", suri.strip())
    mnemonic = parts[0]
    path = suri.replace(mnemonic, "")
    return None if path == "" else path


def detect_derivation(
    address: str, mnemonic: str, langage_code: str, ss58_format: int, crypto_type: int
) -> Optional[str]:
    """
    Detect derivation to add to mnemonic to find address

    :param address: SS58 address
    :param mnemonic: Mnemonic phrase
    :param langage_code: Mnemonic langage code
    :param ss58_format: SS58 Format number
    :param crypto_type: Crypto type constant
    :return:
    """
    # detect root address (empty derivation)
    keypair = Keypair.create_from_mnemonic(
        mnemonic,
        language_code=langage_code,
        ss58_format=ss58_format,
        crypto_type=crypto_type,
    )
    if keypair.ss58_address == address:
        return ""

    # detect member account
    suri = mnemonic + DERIVATION_PATH_MEMBER
    keypair = Keypair.create_from_uri(
        suri,
        language_code=langage_code,
        ss58_format=ss58_format,
        crypto_type=crypto_type,
    )
    if keypair.ss58_address == address:
        return DERIVATION_PATH_MEMBER

    # detect transparent derivation (//[even])
    for account_id in range(2, 100, 2):
        derivation = f"//{account_id}"
        suri = mnemonic + derivation
        keypair = Keypair.create_from_uri(
            suri,
            language_code=langage_code,
            ss58_format=ss58_format,
            crypto_type=crypto_type,
        )
        if keypair.ss58_address == address:
            return derivation

    return None



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, List, Optional

from tikka.domains.entities.smith import Smith
from tikka.libs.keypair import Keypair

if TYPE_CHECKING:
    from interfaces.adapters.network.node.node import NetworkNodeInterface


class NodeSmithsInterface(abc.ABC):
    """
    NodeSmithsInterface class
    """

    def __init__(self, node: NetworkNodeInterface) -> None:
        """
        Use NetworkNodeInterface to request/send smiths information

        :param node: NetworkNodeInterface instance
        :return:
        """
        self.node = node

    @abc.abstractmethod
    def get_smith(self, identity_index: int) -> Optional[Smith]:
        """
        Return Smith instance

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_smiths(self, identity_indice: List[int]) -> List[Optional[Smith]]:
        """
        Return list of Smith instance from list of identity indice

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def invite(self, keypair: Keypair, identity_index: int) -> None:
        """
        The Keypair account invite a member address to be a smith

        :param keypair: Issuer Keypair
        :param identity_index: Recipient identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def accept_invitation(self, keypair: Keypair) -> None:
        """
        The Keypair account accept an invitation to be a smith

        :param keypair: Invited Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def certify(self, keypair: Keypair, identity_index: int) -> None:
        """
        The smith account Keypair certify a member address to be a smith

        :param keypair: Issuer Keypair
        :param identity_index: Recipient identity index
        :return:
        """
        raise NotImplementedError


class NodeSmithsException(Exception):
    """
    NodeSmithsException class
    """
